import numpy as np
from . import SO2
from .utils import *
from rhovee.plot import draw_arrowhead, draw_coord_sys_2d
import matplotlib.pyplot as plt

def plot(ax, se2_mat, size=1.0, label='', style='vec', color=None):
    R,t = to_Rt(se2_mat)
    x_axis = R[:2,0]
    y_axis = R[:2,1]
    #ax.plot(t[0],t[1], 'ro')
    if style == 'vec':
        draw_coord_sys_2d(ax, x_axis, y_axis, t, size, label, color=color)
    elif style == 'dir_x':
        draw_arrowhead(ax, x_axis, t, label=label, color=color)
    elif style == 'dir_y':
        draw_arrowhead(ax, y_axis, t, label=label, color=color)

def plot_trajectory(ax, se2_list, num_plot_poses=5, color='black', size=1):
    num_poses = len(se2_list)
    num_plot_poses = min(num_poses, num_plot_poses)
    idx_poses_to_plot = np.linspace(0,num_poses-1, num_plot_poses).astype(np.uint32)
    for idx in idx_poses_to_plot:
        plot(ax, se2_list[idx], size=size)
    se2_np_arr = np.array(se2_list)
    xs = se2_np_arr[:,0,2]
    ys = se2_np_arr[:,1,2]
    ax.plot(xs,ys)

def identity():
    return np.identity(3)

def tangent_delta(se2_mat, se2_defining_tangent):
    delta = Log(inv(se2_defining_tangent)@se2_mat)
    return delta

def tf_point(se2_mat, point):
    point = np.array(point).flatten()
    R,t = to_Rt(se2_mat)
    return R@point + t
    
def perturb_local(se2_mat, se2_vec):
    return se2_mat@Exp(se2_vec)

def perturb_global(se2_mat, se2_vec):
    return Exp(se2_vec)@se2_mat
    

def manifold_dim():
    return (3,3)

def vector_dim():
    return (3,)

def Rt(R,t):
    assert R.shape == (2,2)
    se2_mat = np.identity(3)
    se2_mat[:2,:2] = R
    se2_mat[:2,2] = t
    return se2_mat

def inv(se2_mat):
    R,t = to_Rt(se2_mat)
    inv_T = np.identity(3)
    inv_T[:2,:2] = R.T
    inv_T[:2,2] = -R.T@t
    return inv_T

def to_Rt(se2_mat):
    R = se2_mat[:2,:2]
    t = se2_mat[2,:2]
    return R,t

def SVDO(se2_mat):
    R,t = to_Rt(se2_mat)
    R = SO2.SVDO(R)
    return Rt(R,t)

def Rz(angle, t=[0,0], unit='rad'):
    t = np.array(t).flatten()
    assert t.shape == (2,)
    angle = angle_as_rad(angle, unit)
    R = SO2.Exp(angle)
    return Rt(R, t)

    


def to_Rt(SE2_mat):
    rot_mat = SE2_mat[:2,:2]
    t = SE2_mat[:2,2]
    return rot_mat, t


def V_helper(theta):
    if np.isclose(theta, 0):
        return np.identity(2)
    V = np.sin(theta)/theta * np.identity(2) + (1-np.cos(theta))/theta*skew2(1)
    return V

def V_inv_helper(theta):
    return np.linalg.inv(V_helper(theta))

def Exp(vec):
    assert vec.shape == (3,)
    SE2_mat = np.identity(3)
    theta = vec[2]
    rho = vec[:2]
    SO2_mat = SO2.Exp(theta)
    SE2_mat[:2,:2] = SO2_mat
    t = V_helper(theta)@rho
    SE2_mat[:2, 2] = t
    return SE2_mat

def Log(SE2_mat):
    assert SE2_mat.shape == (3,3)
    vec = np.zeros(3)
    R,t = to_Rt(SE2_mat)
    theta = SO2.Log(R)
    rho = V_inv_helper(theta)@t
    vec[:2] = rho
    vec[2] = theta
    return vec

def adjoint(SE2_mat):
    ad = np.identity(3)
    R,t = to_Rt(SE2_mat)
    ad[:2,:2] = R
    ad[:2, 2] = -skew2(1)@t
    return ad

def right_jacobian(vec):
    assert vec.shape == (3,)
    rho_1 = vec[0]
    rho_2 = vec[1]
    theta = vec[2]
    if np.allclose(theta, np.zeros(3)):
        return np.identity(3)
    right_jacob = np.zeros((3,3))
    sin_th = np.sin(theta)
    cos_th = np.cos(theta)
    right_jacob[0,0] = sin_th/theta
    right_jacob[1,0] = (cos_th - 1)/theta
    right_jacob[2,0] = 0.0
    right_jacob[0,1] = (1-cos_th)/theta
    right_jacob[1,1] = sin_th/theta
    right_jacob[2,1] = 0.0
    right_jacob[0,2] = (theta*rho_1 - rho_2 + rho_2*cos_th - rho_1*sin_th)/(theta**2)
    right_jacob[1,2] = (rho_1+theta*rho_2-rho_1*cos_th-rho_2*np.sin(theta))/theta**2
    right_jacob[2,2] = 1.0
    return right_jacob

def inv_right_jacobian(vec):
    assert vec.shape == (3,)
    return np.linalg.inv(right_jacobian(vec))

def left_jacobian(vec):
    assert vec.shape == (3,)
    return right_jacobian(-vec)





import numpy as np
import cv2
import matplotlib.pyplot as plt


def draw_centered_line(img_size, angle, line_width):
    # draw a line that goes through the center of the image
    img = np.zeros(img_size)
    center = np.array(img_size) / 2
    x1 = 0
    y1 = int(center[1] - np.tan(angle) * center[0])
    x2 = img.shape[1]
    y2 = int(center[1] + np.tan(angle) * (img.shape[1] - center[0]))
    cv2.line(img, (x1, y1), (x2, y2), (255, 0, 0), 5)
    return img

def weighted_mean_row_index(img, threshold=20, verbose=0):
    assert img.ndim == 2
    img = img.copy()
    img = np.where(img > threshold, img, 0).astype(np.uint8)
    if verbose:
        cv2.imshow('img', img)
        cv2.waitKey(0)
        cv2.destroyAllWindows()
    # find index weighted mean
    indeces = np.arange(img.shape[1])
    mean = np.sum(indeces * img, axis=1) / np.sum(img, axis=1)
    return mean

def row_list_to_points(row_list):
    ys = np.arange(len(row_list))
    xs = row_list
    return np.vstack([xs, ys, np.ones(len(xs))]).T

def draw_homg_line(img, l):
    assert l.shape[0] == 3
    out = img.copy()
    x1 = 0
    y1 = int(-l[2] / l[1])
    x2 = img.shape[1]
    y2 = int(-(l[2] + l[0] * x2) / l[1])
    cv2.line(out, (x1, y1), (x2, y2), (255, 0, 0), 1)
    return out

def fit_homogenous_line(points):
   assert points.shape[1] == 3
   # find m,c using least squares
   points = np.asarray(points)
   n, _ = points.shape
   mean_x = np.mean(points[:,0])
   mean_y = np.mean(points[:,1])
   sum_xy = sum([points[i,0]*points[i,1] for i in range(n)])
   sum_x = sum([points[i,0] for i in range(n)])
   sum_y = sum([points[i,1] for i in range(n)])
   sum_x2 = sum([points[i,0]**2 for i in range(n)])
   slope = (n*sum_xy - sum_x*sum_y) / (n*sum_x2 - sum_x**2)
   intercept = mean_y - slope*mean_x
   return np.array([-slope, 1, -intercept])

def overlap_gray_imgs_rgb(img1, img2):
    assert img1.shape == img2.shape
    rgb_img = np.zeros(img1.shape + (3,))
    rgb_img[..., 0] = img1
    rgb_img[..., 2] = img2
    return rgb_img.astype(np.uint8)





def get_laser_line_as_homg(img, threshold, verbose=0):
    if img.ndim == 3:
        img = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)
    mean = weighted_mean_row_index(img, threshold, verbose)
    points = row_list_to_points(mean)
    l = fit_homogenous_line(points)
    if verbose:
        zero_img = np.zeros(img.shape)
        homg_line_img = draw_homg_line(zero_img, l)
        overlap = overlap_gray_imgs_rgb(img, homg_line_img)
        cv2.imshow('overlap', overlap)
        cv2.waitKey(0)
    return l

def get_essential_matrix(R_12, t_12):
    T = np.eye(4)
    T[:3,:3] = R_12
    T[:3,3] = t_12
    T_inv = np.linalg.inv(T)
    R_21 = T_inv[:3,:3]
    t_21 = T_inv[:3,3]
    E = cross_prod_mat(t_21)@(R_21)
    return E

def homg_points_to_plucker_line(pt1, pt2):
    pt1 = pt1.squeeze()
    pt2 = pt2.squeeze()
    if pt1.shape[0] == 3:
        pt1 = np.append(pt1, 1)
    if pt2.shape[0] == 3:
        pt2 = np.append(pt2, 1)
    l = pt1[3]*pt2[:3] - pt2[3]*pt1[:3]
    l_dash = np.cross(pt1[:3],pt2[:3])
    return l, l_dash

def intersect_plucker_lines(l1, l1_dash, l2, l2_dash):
    n = np.cross(l1, l2)
    v = np.zeros(4)
    v[:3] = np.cross(n, l1)
    v[3] = np.dot(n, l1_dash)
    x = np.zeros(4)
    x[:3] = -v[3]*l2+np.cross(v[:3], l2_dash)
    x[3] = np.dot(v[:3], l2)
    x = x / x[3]
    return x[:3]

def cross_prod_mat(v):
    return np.array([[0, -v[2], v[1]],
                     [v[2], 0, -v[0]],
                     [-v[1], v[0], 0]])

def triangulate_laser_lines(l_img, r_img, thresh, R12, t12, lK, ldc, rK, rdc, verbose=0):
    if l_img.ndim == 3:
        l_img = cv2.cvtColor(l_img, cv2.COLOR_RGB2GRAY)
    if r_img.ndim == 3:
        r_img = cv2.cvtColor(r_img, cv2.COLOR_RGB2GRAY)
    l_img = cv2.undistort(l_img, lK, ldc)
    r_img = cv2.undistort(r_img, rK, rdc)
    l_homg = get_laser_line_as_homg(l_img, thresh, verbose)
    r_homg = get_laser_line_as_homg(r_img, thresh, verbose)
    l_mean_rows = weighted_mean_row_index(l_img, thresh)
    l_points = row_list_to_points(l_mean_rows)
    E = get_essential_matrix(R12, t12)
    F = np.linalg.inv(rK).T @ E @ np.linalg.inv(lK)
    all_pts = []
    for idx,point in enumerate(l_points):
        line = F @ point
        intersect_r = np.cross(r_homg, line)
        intersect_r = intersect_r / intersect_r[2]
        norm_r = np.linalg.inv(rK) @ intersect_r
        norm_r = norm_r / norm_r[2]
        norm_l = R12@norm_r + t12
        line_r, line_r_dash = homg_points_to_plucker_line(norm_l, t12)
        line_l = np.linalg.inv(lK) @ point
        line_l_dash = np.zeros(3)
        intersect = intersect_plucker_lines(line_l, line_l_dash, line_r, line_r_dash)
        all_pts.append(intersect)
    return np.array(all_pts)

def get_proj_planar_homography(R12, t12, u1, K1, K2):
    d = u1[-1]
    n = u1[:3]
    T = np.eye(4)
    T[:3,:3] = R12
    T[:3,3] = t12
    T_inv = np.linalg.inv(T)
    R_inv = T_inv[:3,:3]
    t_inv = T_inv[:3,3]
    H = R_inv - (1/d)*np.outer(t_inv,n)
    H = K2@H@np.linalg.inv(K1)
    H12 = np.linalg.inv(H)
    return H12






if __name__ == '__main__':
    img_path = "/Users/olaals/projects/laser-charuco-calib/stereo-cam-laser-calib/laser-images/test-images/left/img00.png"
    #img = draw_centered_line((500, 500), np.pi / 2.1, 5)
    img = cv2.imread(img_path)
    # blur img
    # cam mat with focal len 500 and center 250,250
    K = np.array([[500, 0, 250], [0, 500, 250], [0, 0, 1]])
    dc = np.array([0, 0, 0, 0])
    # rot mat with -20 deg rot around y axis
    R = np.array([[np.cos(-np.pi/9), 0, np.sin(-np.pi/9)], [0, 1, 0], [-np.sin(-np.pi/9), 0, np.cos(-np.pi/9)]])
    t = np.array([0.5, 0, 0.2])
    pts = triangulate_laser_lines(img, img, 50, R, t, K, dc, K, dc, 1)








from harvesters.core import Harvester
import os
import cv2

class GenicamWrapper():
    def __init__(self, cti_path, device_num, exposure_time=30000, width=2176, height=2176, offset=968, gray=False):
        print("Initializing GenicamWrapper")
        h = Harvester()
        h.add_file(cti_path)
        h.update()
        h.device_info_list[0]
        # set exposure time
        ia = h.create(device_num)
        ia.remote_device.node_map.ExposureTime.value = exposure_time
        ia.remote_device.node_map.Width.value = width
        ia.remote_device.node_map.Height.value = height
        ia.remote_device.node_map.OffsetX.value = offset
        ia.start()
        self.gray = gray
        self.ia = ia
    
    def read(self):
        with self.ia.fetch() as buffer:
            component = buffer.payload.components[0]
            _1d = component.data
            _2d = component.data.reshape(component.height, component.width)
            debayered = cv2.cvtColor(_2d, cv2.COLOR_BAYER_RG2RGB)
            resized = cv2.resize(debayered, (component.width//2, component.width//2))
            if gray:
                resized = cv2.cvtColor(resized, cv2.COLOR_RGB2GRAY)
        return resized

class GenicamStereo():
    def __init__(self, cti_path, d_num_left, d_num_right, exposure_time=30000, width=2176, offset=968, gray=False):
        h = Harvester()
        h.add_file(cti_path)
        h.update()
        h.device_info_list[0]
        ia_left = h.create(d_num_left)
        ia_right = h.create(d_num_right)
        ia_left.remote_device.node_map.ExposureTime.value = exposure_time
        ia_right.remote_device.node_map.ExposureTime.value = exposure_time
        ia_left.remote_device.node_map.Width.value = width
        ia_right.remote_device.node_map.Width.value = width
        ia_left.remote_device.node_map.Height.value = width
        ia_right.remote_device.node_map.Height.value = width
        ia_left.remote_device.node_map.OffsetX.value = offset
        ia_right.remote_device.node_map.OffsetX.value = offset
        ia_left.start()
        ia_right.start()
        self.left = ia_left
        self.right = ia_right
        self.gray = gray

    
    def read(self):
        with self.left.fetch() as buffer:
            component = buffer.payload.components[0]
            _1d = component.data
            _2d = component.data.reshape(component.height, component.width)
            debayered = cv2.cvtColor(_2d, cv2.COLOR_BAYER_RG2RGB)
            resized = cv2.resize(debayered, (component.width//2, component.width//2))
            if self.gray:
                resized = cv2.cvtColor(resized, cv2.COLOR_RGB2GRAY)
            left = resized
        with self.right.fetch() as buffer:
            component = buffer.payload.components[0]
            _1d = component.data
            _2d = component.data.reshape(component.height, component.width)
            debayered = cv2.cvtColor(_2d, cv2.COLOR_BAYER_RG2RGB)
            resized = cv2.resize(debayered, (component.width//2, component.width//2))
            if self.gray:
                resized = cv2.cvtColor(resized, cv2.COLOR_RGB2GRAY)
            right = resized
        return left, right




if __name__ == '__main__':
    cti_path = "/opt/cvb-13.04.005/drivers/genicam/libGevTL.cti.1.2309"
    exposure_time = 60000
    device_num_left = 0
    device_num_right = 1
    cap = GenicamStereo(cti_path, device_num_left, device_num_right, exposure_time, gray=True)
    while True:
        left, right = cap.read()
        cv2.imshow('left', left)
        cv2.imshow('right', right)
        if cv2.waitKey(1) & 0xFF == ord('q'):
            break

    
    





import numpy as np


def create_camera_matrix(fx,fy,u,v):
    return np.array([[fx,0,u],[0,fy,v], [0,0,1]])

def resize_camera_matrix(current_img_size, new_img_size, current_K):
    ratio = new_img_size*1.0/current_img_size
    new_K = current_K*ratio
    new_K[2,2] = 1.0
    return new_K




if __name__ == '__main__':
    pass
    K = create_camera_matrix(640,640,300,300)
    new_K = resize_camera_matrix(1000, 250, K)
    print(K)
    print(new_K)

